/*
** Copyright (c) 1994 The Regents of the University of California.
** All rights reserved.
** 
** Permission is hereby granted, without written agreement and without
** license or royalty fees, to use, copy, modify, and distribute this
** software and its documentation for any purpose, provided that source 
** distributions retain this entire copyright notice and comment.
** 
** Neither the name of the University nor the names of its contributors
** may be used to endorse or promote products derived from this software
** without specific prior written permission.
** 
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
** IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/* --------------------------------
** FILE
**      illu-proc.c
**
** DESCRIPTION
**      Database backend routines for NCSA mosaic-sql gateway and Illustra.
**
**      This interface recognizes the following proc-fmt DEFINEs:
**
**      MI_HOME         sets environment variable MI_HOME
**      MI_SYSPARAMS    sets environment variable MI_SYSPARAMS
**      MI_DATABASE     database name
**      MI_USER         user login
**      MI_PASSWORD     user password
**      MI_SERVER       server name
**      MI_HOST         database host name
**      MI_PORT         tcp/ip port
**      MI_DEBUG        Set to stderr, stdout, or filename
**
**      This has very crude formatting capabilities (should be replaced
**      by something like perl).  A max column width may be set like this:
**
**           DEFINE <column_name> <column_width>
**
**      Word wrap for one column may be set like this:
**
**           DEFINE WORD_WRAP <column_name> 
**
**      Word wrap is limited, hoaky, and hacked.  If you set a wrapped column 
**      width less than 30, it will get set to 30 because this code isn't
**      savvy enough to break midword.  All other columns are truncated at 
**      the column width.
**
** AUTHOR
**      Jean Anderson, Sequoia 2000 Project, May 1994
**      jta@postgres.berkeley.edu, jta@sdsc.edu, jta@ucsd.edu
**      
** BUILD INSTRUCTIONS
**      cc sqlmain.c  sqlutil.c illu-proc.c -I$MI_HOME/h \
**      $MI_HOME/lib/libmi.a -g -lm -o sqlmain
**
** MODS
**      5/28/94  jta  Works with Montage 2.0.9
**      6/07/94  jta  Illustra 2.0.12
**      7/11/94  jta  added minimal word wrap
**      9/30/94  jta  cleanups for release
**
** $Header: /users/vis/jta/www/illu-proc/RCS/illu-proc.c,v 1.17 1994/10/03 17:55:22 jta Exp jta $
** --------------------------------------------------------------------- */

#include "illu-proc.h"

/* execute_sql()
 * Entry point for sqlmain, executes the sql query.
 * Returns 0 on success, MI_ERROR on failure.
 */
int execute_sql (char *sqlstring)
{
     MI_CONNECTION  *conn = NULL;
     char           *query = NULL,
                    *debug_file = NULL;
     int            debug = 0;

     if( (debug_file = (char*) app_getdefine("MI_DEBUG")) != (char *)NULL) 
     {    debug++;
          if(!strncmp(debug_file, "stdout", 6))
               fp=stdout;
          else if(!strncmp(debug_file, "stderr", 6))
               fp=stderr;
          else 
          {    if ( (fp=fopen(debug_file, "w")) == NULL) 
               {    perror("execute_sql: could not open MI_DEBUG file");
                    fp=stdout;
                    fprintf(fp, 
                         "Failed to open file %s; set MI_DEBUG to stdout.\n", 
                         debug_file);
               }
          }
          fprintf(fp, "Values coming into execute_sql:\n");
          fprintf(fp, "execute_sql: original query='%s'\n", sqlstring);
     }

     /* ======== connect to database ======== */
     if ( (conn = init_illustra ()) == (MI_CONNECTION *) NULL)
          return(MI_ERROR);

     /* ======== Rewrite the query ==========
      *  - illustra queries need a terminating semicolon.
      *  - cgi mucks with spaces and some punctuation
      */

     if ( (query = query_rewrite(sqlstring)) == (char *)NULL) 
     {    cleanup(conn, query);
          return(MI_ERROR);
     }

     if(debug)
          fprintf(fp, "execute_sql: actual query='%s'\n", query);

     /* ======== execute query ============== */
     if (mi_exec(conn, query, 0) == MI_ERROR) 
     {    printf("execute_sql: mi_exec() failed.\n");
          cleanup(conn, query);
          return(MI_ERROR);
     }

     /* ======== process query results ====== */
     if (process_results(conn) == MI_ERROR) 
     {    cleanup(conn, query);
          return(MI_ERROR);
     }

     /* ======== clean up and return ======== */
     cleanup(conn, query);
     return(0);
}

/* init_illustra()
 * Initializes illustra connection. 
 * Returns NULL connection on error.
 */
MI_CONNECTION * init_illustra()
{
     char                *routine = "init_illustra",
                         *mi_home = NULL,
                         *mi_sysparams = NULL,
                         *mi_database = NULL,
                         *mi_user = NULL,
                         *mi_password = NULL,
                         *mi_server = NULL,
                         *mi_host = NULL,
                         *mi_port = NULL;
     int                  debug = 0;
     MI_CONNECTION       *conn = NULL;
     MI_CONNECTION_INFO   cinfo;
     MI_DATABASE_INFO     dinfo;

     if ( (char*) app_getdefine("MI_DEBUG") != (char *)NULL)
          debug++;

     /*====================== Set up callback ===============================*/
     if (mi_add_callback(MI_All_Events, (MI_VOID) all_callback, (void *)NULL)
          == MI_ERROR) 
     {    printf("%s: mi_add_callback failed.\n", routine);
          return ((MI_CONNECTION *)NULL);
     }

     /* Set any environment variables */
     if((mi_home = (char*) app_getdefine("MI_HOME")) != (char *)NULL) 
     {    if(debug) fprintf(fp, "mi_home = %s\n", mi_home);
          putenv(mi_home);
     }

     if((mi_sysparams = (char*) app_getdefine("MI_SYSPARAMS")) != (char *)NULL)
     {    if(debug) fprintf(fp, "mi_sysparams = %s\n", mi_sysparams);
          putenv(mi_sysparams);
     }

     /*====================== Connect to Server =============================*/

     /* Set either MI_SERVER, or MI_HOST & MI_PORT (but not all 3) */

     if ((mi_server = (char*) app_getdefine("MI_SERVER")) == (char *)NULL) 
     {    mi_host = (char*) app_getdefine("MI_HOST");
          mi_port = (char*) app_getdefine("MI_PORT");
          if(debug) 
          {    if(mi_host) fprintf(fp, "mi_host = %s, ", mi_host);
               if(mi_port) fprintf(fp, "mi_port = %s\n", mi_port);
          }
     }
     else
          if(debug) fprintf(fp, "mi_server = %s\n", mi_server);

     cinfo.server_name = '\0';
     cinfo.server_host = '\0';
     cinfo.server_port = 0;

     /* If neither server nor host are set, try to get default */

     if ((!mi_server) && (!mi_host)) 
     {    if (mi_get_default_connection_info(&cinfo) == MI_ERROR) 
          {    printf("%s: mi_get_default_connection_info failed.\n", routine);
               return ((MI_CONNECTION *)NULL);
          }
     }
     else 
     {    if(mi_server) 
          {
/*              doesn't work:
 *              cinfo.server_name = mi_server;
 */
                /* This works: */
                if ( (cinfo.server_host = mi_getsysparam(mi_server, "MI_HOST"))
                     == NULL) 
                {    printf("%s: mi_getsysparam failed on MI_HOST.\n", routine);
                     return ((MI_CONNECTION *)NULL);
                }
                if ( (mi_port = mi_getsysparam(mi_server, "MI_PORT")) != NULL)
                     cinfo.server_port = atoi(mi_port);
          }
          else 
          {    cinfo.server_host = mi_host;
               if(mi_port) /* optional */
                    cinfo.server_port = atoi(mi_port);
          }

          if ((mi_set_default_connection_info(&cinfo)) == MI_ERROR)
          {    printf("%s: mi_set_default_connection_info failed.\n", routine);
               return ((MI_CONNECTION *)NULL);
          }
     }

     if(debug) 
     {    fprintf(fp, "%s: Server info: ", routine);
          if(cinfo.server_name)
               fprintf(fp, "server_name=%s, ", cinfo.server_name);
          if(cinfo.server_host)
               fprintf(fp, "server_host=%s, ", cinfo.server_host);
          if(cinfo.server_port)
               fprintf(fp, "server_port=%d\n", cinfo.server_port);
     }

     if ((conn = mi_server_connect(&cinfo)) == (MI_CONNECTION *)NULL) 
     {    printf("%s: mi_server_connect failed.\n", routine);
          return ((MI_CONNECTION *)NULL);
     }

     /*===================== Login to Database ============================*/

     if (mi_get_default_database_info(&dinfo) == MI_ERROR) 
     {    printf("%s: mi_get_default_database_info failed.\n", routine);
          return ((MI_CONNECTION *)NULL);
     }

     /* override defaults */
     if ((mi_database = (char*) app_getdefine("MI_DATABASE")) != (char *)NULL)
           dinfo.database_name = mi_database;
     if ((mi_user = (char*) app_getdefine("MI_USER")) != (char *)NULL)
           dinfo.user_name = mi_user;
     if ((mi_password = (char*) app_getdefine("MI_PASSWORD")) != (char *)NULL)
           dinfo.password = mi_password;

     /* Don't output password :-) ...  */
     if(debug) 
     {    fprintf(fp, "%s: Server info: ", routine);
          fprintf(fp, "database_name=%s, user_name=%s\n",
               dinfo.database_name, dinfo.user_name);
     }

     if (mi_login(conn, &dinfo) == MI_ERROR) 
     {    printf("%s: mi_login failed.\n", routine);
          return ((MI_CONNECTION *)NULL);
     }

     return(conn);
}


/* process_results()
 * Processes query results. 
 * Returns 0 on success, MI_ERROR on error.
 */
int process_results(MI_CONNECTION *conn)
{
     char *r = "process_results";
     int   ret = 0;

     printf("<P><H2> RESULTS: </H2><P>" );
     printf("<PRE>\n"); 

     while ((ret = mi_get_result(conn)) != MI_NO_MORE_RESULTS) 
     {    switch (ret) 
          {    case MI_ERROR:
                    printf("%s: mi_get_result() failed.\n", r);
                    return(MI_ERROR);
               case MI_DDL:                /* FALLTHROUGH */
               case MI_DML:
                    break;
               case MI_ROWS:
                    if(print_results(conn) == MI_ERROR) 
                         return(MI_ERROR);
                    break;
               default:
                    printf("%s: Unexpected mi_get_result() result\n", r);
                    return(MI_ERROR);
          }
     }

     printf("</PRE>\n"); 
     return(0);
}

/* print_results()
 * Outputs results. 
 * Returns 0 on success, MI_ERROR on error.
 */
int print_results(MI_CONNECTION *conn)
{
     char        *r = "print_results", 
                 *ww_col = NULL,        /* name of the wrapped column        */
                **ww_ptr = NULL,        /* pointers into the wrapped column  */
                  blank[PAD_SIZE];      /* for between fields and word wrap  */
     int          error = 0,            /* status from mi_next_row()         */
                  entering = 1,         /* get column info                   */
                  ncols = 0,            /* mi_column_count()                 */
                  i,                    /* loop counter for col processing   */
                  row_count = 0,        /* keeps track of how many rows      */
                  ww = -1,              /* column number to be wrapped       */
                  ww_array = 0,         /* number of lines in wrapped column */
                  debug = 0, 
                  line_size = 0;        /* debug WinMosaic/Daytona problem   */
     size_t       ww_num = 0;           /* allocated size of ww_ptr          */
     MI_ROW      *row = NULL;
     COLUMN_INFO *column_data = NULL;

     memset(blank, ' ', sizeof(blank));
     blank[sizeof(blank) - 1] = '\0';

     if ( (char*) app_getdefine("MI_DEBUG") != (char *)NULL)
          debug++;

     /* If there is a word_wrapped column, allocate the pointers that 
      * will make the long text string look like an array of strings.
      */
     if((ww_col = (char*) app_getdefine("WORD_WRAP")) != NULL) 
     {    if((alloc_ww_buff(&ww_ptr, &ww_num)) == MI_ERROR)
               return(MI_ERROR);
          ww = 0; /* initialize to first column */
     }

     while ((row = mi_next_row(conn, &error)) != (MI_ROW *)NULL) 
     {    
          /* =================== Get header info ============================
           * Note: This code is simple. Column info can change in Illustra
           * as you traverse an inheritance hierarchy or process a composite
           * column type.  $MI_HOME/examples/documentation/jagged.c shows 
           * how to handle this kind of data and will be worked into this 
           * code eventually.
           */
          if (entering)
          {    if((get_header_info(row, &ncols, &column_data, ww_col, &ww))
                    == MI_ERROR)
               {    cleanup_results(ww_ptr, column_data, ncols);
                    return(MI_ERROR);
               }
               entering = 0;
               if(debug)
               {    for (i = 0; i < ncols; i++)
                    {    printf("%s: %-35s = %d + %d\n", r,
                              column_data[i].col_name, 
                              column_data[i].col_width,
                              FIELD_SEP_SIZE);
                         line_size += column_data[i].col_width + FIELD_SEP_SIZE;
                    }
                    printf("%s: line size of result set (%d columns) = %d\n", 
                         r, ncols, line_size);
               }
          }
          if(row_count % 20 == 0)
               print_header(column_data, ncols);

          /* ======== Get column values for current row ===== */
          if((get_column_vals(row, ncols, column_data)) == MI_ERROR)
          {    cleanup_results(ww_ptr, column_data, ncols);
               return(MI_ERROR);
          }

          /* ======== Output values for current row ============ */
          for (i = 0; i < ncols; i++)
          {    if ((i == ww) 
                && (column_data[i].col_data_len > column_data[i].col_width))
               {    ww_array=set_ww_ptr(column_data[i].col_data, 
                         column_data[i].col_width, &ww_ptr,&ww_num);
                    if(ww_array < 0)
                    {    printf("set_ww_ptr failed\n");
                         cleanup_results(ww_ptr, column_data, ncols);
                         return(MI_ERROR);
                    }
                    printf("%-*.*s", 
                         column_data[i].col_width,
                         column_data[i].col_width,
                         (char *)ww_ptr[0]);
               }
               else
               {    printf("%-*.*s", 
                         column_data[i].col_width,
                         column_data[i].col_width,
                         column_data[i].col_data);
                    if(i == ww) /* this is a wrapped col with nothing more */
                         ww_array = 0;
               }
               printf("%-*.*s", FIELD_SEP_SIZE, FIELD_SEP_SIZE, blank);
          }
          printf("\n");
          
          /* ======== Output Wrapped column ============ */

          if ((ww >= 0) && (ww_array > 0))
          {    for (i = 1; i < ww_array; i++) /* we already ouput first row */
               {    printf("%-*.*s",
                         column_data[ww].col_offset,
                         column_data[ww].col_offset,
                         blank);
                    printf("%-*.*s\n", 
                         column_data[ww].col_width,
                         column_data[ww].col_width,
                         (char *)ww_ptr[i]);
               }
          } 
          row_count++;
     } /* End process each row */

     cleanup_results(ww_ptr, column_data, ncols);

     if (error == MI_ERROR) 
     {    printf("%s: mi_next_row failed", r);
          return(MI_ERROR);
     } 

     if(row_count)
          printf("\n\nQuery returned %d %s.\n", 
               row_count, (row_count == 1) ? "row" : "rows");
     else
          printf("\nQuery returned no rows.\n");

     return(0);
}

/* get_header_info()
 * Allocations column_data and populates it with information for each column 
 * in the query results.
 * Returns 0 on success, MI_ERROR on error.
 */
int get_header_info(MI_ROW *row, int *ncols, COLUMN_INFO **column_data, 
    char *ww_col, int *ww)
{
     char         *r = "get_header_info", *p, *tmp_col;
     int           i;
     MI_ROW_DESC  *desc = NULL;

     if( (desc = mi_get_row_desc(row)) == (MI_ROW_DESC *)NULL) 
     {    printf("%s: mi_get_row_desc failed.", r);
          return(MI_ERROR);
     }

     if( (*ncols = mi_column_count(desc)) == MI_ERROR) 
     {    printf("%s: mi_column_count failed.", r);
          return(MI_ERROR);
     }

     if((*column_data = (COLUMN_INFO *) mi_alloc(*ncols * sizeof(COLUMN_INFO)))
          == (COLUMN_INFO *)NULL)
     {    printf("%s: mi_alloc failed.", r);
          return(MI_ERROR);
     }

     for (i = 0; i < *ncols; i++) 
     {    tmp_col = mi_column_name(desc, i);
          if(((*column_data)[i].col_name = 
               (char *)mi_alloc (strlen(tmp_col)+1 * sizeof(char))) 
                    == (char *) NULL)
          {    printf("%s: mi_alloc failed.", r);
               mi_free((char *)*column_data);
               return(MI_ERROR);
          }
          strcpy((*column_data)[i].col_name, tmp_col); 

          /* Does this column have a max width set? */
          if((p=(char*) app_getdefine((*column_data)[i].col_name)) == (char *)NULL)
               (*column_data)[i].col_width = DEFAULT_COL_SIZE;
          else
               (*column_data)[i].col_width = atoi(p);

          /* Is this column word wrapped? 
           * Nasty hack! Word wrap handling in this code breaks on a space 
           * and isn't smart enough to handle narrow columns. So if this is 
           * a word wrapped column, make sure it is wide enough. How wide 
           * is wide enough? Shoot, I dunno. :-)
           */
          if(ww_col) 
          {    if(!strcmp(ww_col, (*column_data)[i].col_name))
               {    *ww = i;
                    if( (*column_data)[i].col_width < 30 )
                         (*column_data)[i].col_width = 30;
               }
          }

          /* offsets are used for word_wrap */
          if(i == 0) 
               (*column_data)[i].col_offset = 0;
          else
               (*column_data)[i].col_offset = 
                    (*column_data)[i-1].col_offset +
                    (*column_data)[i-1].col_width + 
                    FIELD_SEP_SIZE;

          /* initialize the rest */
          (*column_data)[i].col_data = NULL;
          (*column_data)[i].col_data_len = 0;
     }
     return(0);
}

/* get_column_vals()
 * Gets column values for each row in the result set.
 * Returns 0 on success, MI_ERROR on error.
 */
int get_column_vals(MI_ROW *row, int ncols, COLUMN_INFO *column_data)
{
     char *r = "get_column_vals";
     int   i;

     for (i = 0; i < ncols; i++) 
     {    char *col_val = NULL; 
          int  col_len = 0;
          switch (mi_value(row, i, &col_val, &col_len)) 
          {  case MI_ERROR:
                printf("%s: mi_value failed", r);
                return(MI_ERROR);
             case MI_NULL_VALUE:
                col_val = "NULL";
                break;
             case MI_NORMAL_VALUE:
                break;
             default:
                printf("%s: Unexpected column type found.", r);
                return(MI_ERROR);
          }
          column_data[i].col_data_len = col_len;
          column_data[i].col_data = col_val;
          column_data[i].last_offset = 0; /* initialize for word wrap */
     }
     return(0);
}

/* print_header()
 * Given the column_data, outputs column headings.
 * Returns 0 on success, MI_ERROR on error.
 */
void print_header(COLUMN_INFO *column_data, int ncols)
{
     int    i;
     static int    first=0;
     static char   underline[PAD_SIZE],  /* underlines column headings       */
                   blank[PAD_SIZE];      /* for between fields and word wrap */

     if (first == 0)
     {    memset(blank, ' ', sizeof(blank));
          blank[sizeof(blank) - 1] = '\0';
          memset(underline, '-', sizeof(underline));
          underline[sizeof(underline) - 1] = '\0';
          first++;
     }

     for (i = 0; i < ncols; i++)            /* === overline column names === */
     {    printf("%-*.*s", 
               column_data[i].col_width, column_data[i].col_width, underline);
          printf("%-*.*s", FIELD_SEP_SIZE, FIELD_SEP_SIZE, blank);
     }
     printf("\n");

     for(i=0; i < ncols; i++)               /* === output column names ===== */
     {    printf("%-*.*s", 
               column_data[i].col_width, 
               column_data[i].col_width, 
               column_data[i].col_name);
          printf("%-*.*s", FIELD_SEP_SIZE, FIELD_SEP_SIZE, blank);
     }
     printf("\n");

     for (i = 0; i < ncols; i++)            /* === underline column names === */
     {    printf("%-*.*s", 
               column_data[i].col_width, column_data[i].col_width, underline);
          printf("%-*.*s", FIELD_SEP_SIZE, FIELD_SEP_SIZE, blank);
     }
     printf("\n");
     return;
}

/* allow_ww_buff()
 * Allocates pointers that are used for pointing into a wrapped column
 * so it can be treated as an array of strings.
 * Returns 0 on success, MI_ERROR on error.
 */
int alloc_ww_buff(char ***buff, size_t *ww_num)
{
     char  **rptr = NULL;
     size_t  alloc_size = WRAP_LINES + *ww_num;

     if((rptr=(char **) mi_alloc(alloc_size * sizeof(char *))) == NULL) 
     {    printf("mi_alloc failed\n");
          return(MI_ERROR);
     }

     if (*ww_num > 0)
     {    memcpy(rptr, *buff, (*ww_num * sizeof(char *)));
          mi_free((char *)*buff);
     }

     (*buff)=rptr;
     *ww_num = alloc_size;
     return(0);
}

/* set_ww_ptr()
 * Given a long text of string, this sets pointers into the string so
 * it can be treated like an array of strings. Algorithm is rudimentary;
 * it simply backtracks to the first space.
 * Returns 0 on success, MI_ERROR on error.
 */
int set_ww_ptr(char *in, int col_width, char ***out, size_t *ww_num)
{
     char  *p = NULL;
     int    done = 0, 
            rows = 0;

     p = in;
     (*out)[rows] = p;

     while(done == 0) 
     {    if(!(p) || (strlen(p) <= col_width))
               done++;
          else 
          {    rows++;
               if(rows == *ww_num)      /* Did we allocate enough pointers?  */
                    if((alloc_ww_buff(out, ww_num)) == MI_ERROR)
                        return(MI_ERROR);

               p += col_width;
               while (!isspace(*p))     /* back up until you hit a space */
                    p--;
               *p = '\0';               /* make it null,   */
               p++;                     /* move past,      */
               (*out)[rows] = p;        /* and set pointer */
          }
     }
     return(rows + 1);
}

/* all_callback()
 * Handles all illustra messages.
 */
void all_callback(MI_EVENT_TYPE type, MI_CONNECTION *conn, void *s1, void *s2)
{
     int       ecode;
     char      buf[8192], *s;

     switch (type)
     {    case MI_Exception:
               mi_errmsg(s1, buf, 8192);
               printf("%s\n", buf);
               break;
          case MI_Client_Library_Error:
               ecode = mi_error_level(s1);
               switch (ecode)
               {    case MI_LIB_BADARG:   s = "MI_LIB_BADARG";   break;
                    case MI_LIB_USAGE:    s = "MI_LIB_USAGE";    break;
                    case MI_LIB_INTERR:   s = "MI_LIB_INTERR";   break;
                    case MI_LIB_NOIMP:    s = "MI_LIB_NOIMP";    break;
                    case MI_LIB_DROPCONN: s = "MI_LIB_DROPCONN"; break;
                    case MI_LIB_BADSERV:  s = "MI_LIB_BADSERV";  break;
                    default:              s = "UNKNOWN";         break;
               }
               mi_errmsg(s1, buf, 100);
               printf("%s: %s\n", s, buf);
               break;
          case MI_Alerter_Fire_Msg:
               printf("Alerter %s %s\n",
                    mi_alert_name(s1),
                    mi_alert_status(s1) == MI_ALERTER_DROPPED ?
                         "dropped" : "fired");
               break;
          default:
               printf("Caught an unexpected event type\n");
               break;
     }
     return;
}

/* query_rewrite()
 * Takes the incoming sql statement and modifies it as follows:
 *   - replaces + with a space
 *   - converts cgi hex codes back to original
 *   - makes sure that query has a semicolon terminator.
 * Returns new query on success, NULL on failure.
 */
char * query_rewrite(char *str)
{
     char     *r = "query_rewrite",
              tmp_char[4], *s_copy = NULL, *q = NULL, *p = NULL;
     size_t   qsize = 0;

     qsize = strlen(str) + (size_t) 2;   /* max sqlstring + semicolon + NULL */

     if( (q = (char *) calloc(qsize, sizeof(char))) == NULL) 
     {    printf("%s: calloc failed!", r);
          return((char *)NULL);
     }

     /* make a copy of the query -- strtok will stomp on it */
     if( (s_copy = (char *) calloc(qsize, sizeof(char))) == NULL) 
     {    printf("%s: calloc failed!", r);
          return((char *)NULL);
     }
     strcpy(s_copy, str);

     for(p=s_copy; *p; p++)         /* pass #1 changes '+' to space */
         if(*p == '+')
              *p = ' ';

     p = strtok(s_copy, "%");       /* pass #2 translates hex codes */
     if(p) 
     {    strcpy(q, p);
          while ( (p = strtok((char *)NULL, "%")) != NULL) 
          {    if(p) 
               {    memset(tmp_char, '\0', sizeof(tmp_char));
                    decode_cgi_hex(p, tmp_char);
                    strcat(q, tmp_char);
                    p += 2;        /* it's a two digit code */
                    strcat(q, p);
               }
          }
     }
     else
          strcpy(q, str);
     free(s_copy);

     p = q + strlen(q)-1;          /* pass #3 makes sure it ends with ';' */
     while (*p == ' ')             /* back up over any trailing blanks */
          p--;
     if(*p != ';')
     {    p++;                     /* we alloc'd extra for a semicolon */
          *p = ';';    
     }
     return(q);
}

/* decode_cgi_hex()
 * Unmuck the mucking of the Common Gateway Interface (CGI).
 */

struct cgi_hex_codes
{
     char *cgi_code;
     char *result;
} CGI_Hex_Codes[] = { 
     "21", "!",
     "22", "\"",
     "23", "#",
     "24", "$",
     "25", "%",
     "26", "&",
     "27", "\'",
     "28", "(",
     "29", ")",
     "2B", "+", 
     "2C", ",",
     "2F", "/",
     "3A", ":",
     "3B", ";",
     "3C", "<",
     "3D", "=",
     "3E", ">",
     "3F", "?",
     "5B", "[",
     "5C", "\\",
     "5D", "]",
     "5E", "^",
     "60", "`",
     "7B", "{",
     "7C", "|",
     "7D", "}",
     "7E", "~"
};

#define Hex_CodesN ((sizeof(CGI_Hex_Codes)/sizeof(struct cgi_hex_codes)))

void decode_cgi_hex(char *str, char *result)
{
     int  i, found_it = 0;

     for(i=0; i < Hex_CodesN && !found_it; i++) 
     {    if(!strncmp(str, CGI_Hex_Codes[i].cgi_code, (size_t) 2)) 
          {    result[0] = CGI_Hex_Codes[i].result[0];
               found_it++;
          }
     }

     if(found_it == 0) /* put it back the way it was */
     {    result[0] = '%';
          strncat(result, str, (size_t) 2);
     }
     return;
}

void cleanup(MI_CONNECTION *conn, char *query)
{
     if(conn)
          mi_close(conn);
     if(query)
          free(query);
     return;
}

void cleanup_results(char **ww_ptr, COLUMN_INFO *column_data, int ncols)
{
     if(column_data)
          free_header_info(column_data, ncols);

     if (ww_ptr)
          mi_free((char *)ww_ptr);
     return;
}

void free_header_info(COLUMN_INFO *column_data, int ncols)
{
     int i;
     for (i = 0; i < ncols; i++)
          mi_free(column_data[i].col_name);
     mi_free((char *)column_data);
     column_data = NULL;
     return;
}
